#!/usr/bin/env python
# -*- coding: utf-8 -*-

from spiegelib.core.audio_buffer import AudioBuffer
from spiegelib.core.dataset_generator import DatasetGenerator
from spiegelib.core.sound_match import SoundMatch

import spiegelib.evaluation
import spiegelib.estimator
import spiegelib.features
import spiegelib.synth



#!/usr/bin/env python
"""
Custom highway layer implemented in Keras.

Based on implementation by Kadam Parikh: https://github.com/ParikhKadam/Highway-Layer-Keras
"""
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import layers
from tensorflow.keras.initializers import Constant


class HighwayLayer(layers.Layer):
    """
    Args:
        activation (optional): Activation function of name of built-in activation function.
        transform_gate_bias (optional): Initializer for transform gate bias vector
        transform_dropout (float, optional): Dropout rate between 0 and 1 for transform gate.
            Defaults to None and no dropout is applied.
        activity_regularizer: Optional activity regulizer applied to the regular dense layer
        kwargs: Optional keyword arguments passed to ``tf.keras.layers.Layer``. See
            `TensorFlow docs <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer>`_.
    """

    def __init__(
        self,
        activation='relu',
        transform_gate_bias=-1,
        transform_dropout=None,
        activity_regularizer=None,
        **kwargs
    ):
        """
        Constructor
        """
        self.activation = activation
        self.transform_gate_bias = transform_gate_bias
        self.transform_dropout = transform_dropout
        self.activity_regularizer = activity_regularizer
        super(HighwayLayer, self).__init__(**kwargs)


    def build(self, input_shape):

        # Create a trainable weight variable for this layer.
        dim = input_shape[-1]
        transform_gate_bias_initializer = Constant(self.transform_gate_bias)
        input_shape_dense_1 = input_shape[-1]

        # Transform Gate
        self.dense_1 = layers.Dense(
            name='Dense_1',
            units=dim,
            bias_initializer=transform_gate_bias_initializer
        )

        # Regular Dense Layer
        self.dense_2 = layers.Dense(
            name='Dense_2',
            units=dim,
            activity_regularizer=self.activity_regularizer
        )

        super(HighwayLayer, self).build(input_shape)

    def call(self, x):
        dim = K.int_shape(x)[-1]

        # Transform gate operation
        transform_gate = self.dense_1(x)
        transform_gate = layers.Activation("sigmoid")(transform_gate)
        if self.transform_dropout:
            transform_gate = layers.Dropout(self.transform_dropout)(transform_gate)

        # Carry gate operation - determine how much to feedforward
        carry_gate = layers.Lambda(lambda x: 1.0 - x, output_shape=(dim,))(transform_gate)

        transformed_data = self.dense_2(x)
        transformed_data = layers.Activation(self.activation)(transformed_data)
        transformed_gated = layers.Multiply()([transform_gate, transformed_data])

        identity_gated = layers.Multiply()([carry_gate, x])
        value = layers.Add()([transformed_gated, identity_gated])
        return value

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config['activation'] = self.activation
        config['transform_gate_bias'] = self.transform_gate_bias
        return config



#!/usr/bin/env python

"""
Init file for features
"""
from .data_scaler_base import DataScalerBase
from .standard_scaler import StandardScaler

from .features_base import FeaturesBase
from .fft import FFT
from .mfcc import MFCC
from .spectral_summarized import SpectralSummarized
from .stft import STFT



#!/usr/bin/env python

"""
Audio feature extraction utilities
"""

import numpy as np

spectrum_types = ['complex', 'magnitude', 'power', 'magnitude_phase', 'power_phase']
def convert_spectrum(spectrum, type, dtype=np.float32, complex_dtype=np.complex64):
    """
    Convert a raw spectrum to magnitue, power, magnitude with phase, or power
    with phase

    :param spectrum: raw audio spectrum with real/imag values to convert
    :type spectrum: np.ndarray
    :param type: type of conversion to apply ('magnitude', 'power', 'magnitude_phase',
        'power_phase')
    :type type: str
    :param dtype: number type of regular (non-complex) numbers to return, defaults to np.float32
    :type dtype: np number type, optional
    :param complex_dtype: complex number type, defaults to np.complex64
    :type complex_dtype: np complex number type, optional
    :returns: converted spectrum
    :rtype: np.ndarray
    """

    if not type in spectrum_types:
        raise ValueError("type must be one of %s, recevied %s" % (spectrum_types, type))

    # Convert to desired output format and data type
    if type == 'magnitude':
        return np.array(np.abs(spectrum), dtype=dtype)

    elif type == 'power':
        return np.array(np.abs(spectrum)**2, dtype=dtype)

    elif type == 'magnitude_phase':
        magnitude = np.array(np.abs(spectrum), dtype=dtype)
        phase = np.array(np.angle(spectrum), dtype=dtype)
        return np.stack((magnitude, phase), axis=-1)

    elif type == 'power_phase':
        power = np.array(np.abs(spectrum)**2, dtype=dtype)
        phase = np.array(np.angle(spectrum), dtype=dtype)
        return np.stack((power, phase), axis=-1)

    # If we get here, then the type is complex, so return a complex array
    # with an updated data type
    return np.array(spectrum, dtype=complex_dtype)



#!/usr/bin/env python
"""
Abstract Base Class for Data Scalers
"""

from abc import ABC, abstractmethod

class DataScalerBase(ABC):

    @abstractmethod
    def fit(self, data, axis=None):
        """
        Fit parameters for data scaling. Abstract, must be implemented.

        :param data: data to use for fitting scaling parameters
        :type data: np.ndarray
        :param axis: axis or axes to use for calculating scaling parameters on.
        :type axis: int, tuple
        """
        raise NotImplementedError


    @abstractmethod
    def transform(self, data):
        """
        Perform scaling on data

        :param data: data to scale
        :type data: np.array
        :returns: sacled data
        :rtype: np.ndarray
        """
        raise NotImplementedError


    def fit_transform(self, data, axis=None):
        """
        Fit scaling parameters and then scale data

        :param data: data to fit scaling parameters to and then transform
        :type data: np.array
        :returns: scaled data
        :rtype: np.ndarray
        """

        self.fit(data, axis)
        return self.transform(data)



#!/usr/bin/env python
"""
Contains utility functions and classes used in library.
"""


import re
import numpy as np
import json


#: Selection of spectrum outputs
spectrum_types = ['complex', 'magnitude', 'power', 'magnitude_phase', 'power_phase']

def convert_spectrum(spectrum, type, dtype=np.float32, complex_dtype=np.complex64):
    """
    Convert a raw spectrum to magnitue, power, magnitude with phase, or power
    with phase. If output is magnitude or power, the output array shape is the
    same as the input. If either of the options with phase are selected, the output is
    the same shape but each element is replaced with a 2-element vector containing
    the magnitude or power in the first element and the phase in the second.

    Args:
        spectrum (np.ndarray): array of complex values
        type (str): type of conversion to apply ('magnitude', 'power', 'magnitude_phase', 'power_phase')
        dtype (np number type, optional): number type of regular (non-complex) numbers to return, defaults to np.float32
        complex_dtype (np complex type, optional): complex number type, defaults to np.complex64

    Returns:
        np.ndarray: converted spectrum
    """

    if not type in spectrum_types:
        raise ValueError("type must be one of %s, recevied %s" % (spectrum_types, type))

    # Convert to desired output format and data type
    if type == 'magnitude':
        return np.array(np.abs(spectrum), dtype=dtype)

    elif type == 'power':
        return np.array(np.abs(spectrum)**2, dtype=dtype)

    elif type == 'magnitude_phase':
        magnitude = np.array(np.abs(spectrum), dtype=dtype)
        phase = np.array(np.angle(spectrum), dtype=dtype)
        return np.stack((magnitude, phase), axis=-1)

    elif type == 'power_phase':
        power = np.array(np.abs(spectrum)**2, dtype=dtype)
        phase = np.array(np.angle(spectrum), dtype=dtype)
        return np.stack((power, phase), axis=-1)

    # If we get here, then the type is complex, so return a complex array
    # with an updated data type
    return np.array(spectrum, dtype=complex_dtype)


def atoi(text):
    """
    Convert string to integer if it is an integer, otherwise keep it a str

    Args:
        text (str)

    Returns:
        int or str
    """

    return int(text) if text.isdigit() else text

def natural_keys(text):
    """
    Keys for sorting text in natural ordering. For text containing numbers,
    makes sure numbers are sorted in a natural way.

    Usage:
        alist.sort(key=natural_keys) sorts in human order

    From `stackoverflow <https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside>`_

    Args:
        text (str): text to extract natural keys from

    Returns:
        list: List of text and integers separated from text to use as the key argument in a sort algorithm.
    """
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]



class NumpyNumberEncoder(json.JSONEncoder):
    """
    Encodes Numpy number values into values that can be saved as JSON files.

    Usage::

        with open(path, 'w') as fp:
            json.dump(data, fp, indent=True, cls=NumpyNumberEncoder)
    """

    def default(self, obj):
        if type(obj).__module__ == np.__name__:
            try:
                return float(obj)
            except TypeError:
                pass
        return json.JSONEncoder.default(self, obj)



#!/usr/bin/env python
"""
Abstract Base Class for Estimating Synthesizer Parameters using TensorFlow.

Wraps TensorFlow library calls and provides functionality for training and making
synthesizer parameter predictions. Inheriting classes must implement the ``build_model()``
method, which is automatically called upon construction and sets up the neural network
model.

Examples
^^^^^^^^

Here is an example of extending ``TFEstimatorBase`` to create a simple multi-layer
perceptron network::

        import tensorflow as tf
        from tensorflow.keras import layers
        import spiegelib.estimators import TFEstimatorBase

        class MLP(TFEstimatorBase):

            def __init__(self, input_shape, num_outputs, **kwargs):

                # Call TFEstimatorBase constructor
                super().__init__(input_shape, num_outputs, **kwargs)


            def build_model(self):

                # Model must be defined in the model attribute
                self.model = tf.keras.Sequential()
                self.model.add(layers.Dense(50, input_shape=self.input_shape, activation='relu'))
                self.model.add(layers.Dense(40, activation='relu'))
                self.model.add(layers.Dense(30, activation='relu'))
                self.model.add(layers.Dense(self.num_outputs))

                self.model.compile(
                    optimizer=tf.optimizers.Adam(),
                    loss=TFEstimatorBase.rms_error,
                    metrics=['accuracy']
                )

For a detailed example of running a synthesizer sound matching experiment using
deep learning models like this, see the :ref:`FM Sound Match Example <fm_sound_match>`.

"""

import os
from abc import abstractmethod
import datetime
import numpy as np
import tensorflow as tf

from spiegelib.estimator.estimator_base import EstimatorBase
from spiegelib.estimator.tf_epoch_logger import TFEpochLogger
from spiegelib.estimator.highway_layer import HighwayLayer

class TFEstimatorBase(EstimatorBase):
    """
    Args:
        input_shape (tuple, optional): Shape of matrix that will be passed to model input
        num_outputes (int, optional): Number of outputs the model has. If estimating
            synthesizer parameters this will typically be the number of parameters.
        weights_path (string, optional): If given, model weights will be loaded from this file
        callbacks (list, optional): A list of callbacks to be passed into model fit method

    Attributes:
        model (``tf.keras.Model``): Attribute for the model, see `TensorFlow docs <https://www.tensorflow.org/api_docs/python/tf/keras/Model>`__
    """

    def __init__(self, input_shape=None, num_outputs=None,
                 weights_path = "", callbacks=[]):
        """
        Constructor
        """

        super().__init__()

        self.input_shape = input_shape
        self.num_outputs = num_outputs

        # Construct the model
        self.model = None
        self.build_model()

        # Datasets
        self.train_data = None
        self.test_data = None

        # Loggers
        if isinstance(callbacks, list):
            self.callbacks = callbacks
        else:
            raise Exception('loggers argument must be of type list, '
                            'received %s' % type(loggers))

        # Attempt to load model weights if provided
        if weights_path:
            self.load_weights(weights_path)


    @abstractmethod
    def build_model(self):
        """
        Abstract method that should contain the model definition when implemented
        """
        pass


    def add_training_data(self, input, output, batch_size=64,
                          shuffle_size=None):
        """
        Create a `tf.data.Dataset <https://www.tensorflow.org/api_docs/python/tf/data/Dataset>`_
        from training data, and shuffles / batches data for training. Stores
        results in the ``train_data`` attribute.

        Args:
            input (np.ndarray): training data tensor (ex, audio features)
            output (np.ndarray): ground truth data (ex, parameter values)
            batch_size: (int, optional): If provided, will batch data into batches of
                this size. None or 0 will be no batching. Defaults to 64.
            shuffle_size (int, optional): If provided will shuffle data with a buffer
                of this size. None or 0 corresponds to no shuffling. Defaults to None.
        """

        if not input.shape[1:] == self.input_shape:
            raise Exception('Expected training data to have shape %s, got %s' \
                            % (input.shape[1:], self.input_shape))

        if not output.shape[-1] == self.num_outputs:
            raise Exception('Expected training validation data to have shape '
                            '%s, got %s' % (output.shape[-1], self.num_outputs))

        self.train_data = tf.data.Dataset.from_tensor_slices((input, output))

        if shuffle_size:
            self.train_data = self.train_data.shuffle(shuffle_size)

        if batch_size:
            self.train_data = self.train_data.batch(batch_size)


    def add_testing_data(self, input, output, batch_size=64):
        """
        Create a `tf.data.Dataset <https://www.tensorflow.org/api_docs/python/tf/data/Dataset>`_
        and optionally batches for validation. Stores results in the ``test_data`` attribute.

        Args:
            input (np.ndarray): validation data (ex, audio features)
            output (np.ndarray): ground truth data (ex, parameter values)
            batch_size: (int, optional): If provided, will batch data into batches of
                this size. None or 0 will be no batching. Defaults to 64.
        """

        if not input.shape[1:] == self.input_shape:
            raise Exception('Expected training data to have shape %s, got %s' \
                            % (input.shape[1:], self.input_shape))

        if not output.shape[-1] == self.num_outputs:
            raise Exception('Expected training validation data to have shape %s, '
                             'got %s' % (output.shape[-1], self.num_outputs))

        self.test_data = tf.data.Dataset.from_tensor_slices((input, output))

        if batch_size:
            self.test_data = self.test_data.batch(batch_size)


    def fit(self, epochs=1, callbacks=[], **kwargs):
        """
        Train model on for a fixed number of epochs on training data and validation
        data if it has been added to this estimator

        Args:
            epochs (int, optional): Number of epocs to train model on.
            callbacks (list, optional): List of callback functions for training.
            kwargs: Keyword args passed to model fit method. See
                `Tensflow Docs <https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit>`_.
        """

        # Check for callbacks in k
        if not isinstance(callbacks, list):
            raise Exception('Callbacks must be a list, received %s' \
                            % type(callbacks))

        # Add callbacks
        if self.callbacks:
            for callback in self.callbacks:
                callbacks.append(callback)

        # Train model
        self.model.fit(
            self.train_data,
            epochs = epochs,
            validation_data = self.test_data,
            callbacks=callbacks,
            **kwargs
        )

    def predict(self, input):
        """
        Run prediction on input

        Arg (np.ndarray): matrix of input data to run predictions on. Can a single
            instance or a batch
        """

        # If this is a single instance we need to wrap in an np.array
        is_single_input = False
        if input.shape == self.input_shape:
            is_single_input = True
            input = np.array([input])
        else:
            raise Exception('Input data has incorrect shape, expected %s, '
                             'got %s' % (self.input_shape, input.shape))

        prediction = self.model.predict(input)
        return prediction[0] if is_single_input else prediction


    def load_weights(self, filepath, **kwargs):
        """
        Load model weights from H5 or TensorFlow file

        Args:
            filepath (string): location of saved model weights
            kwargs: optional keyword arguments passed to tf load_weights
                methods, see  `TensorFlow Docs <https://www.tensorflow.org/api_docs/python/tf/keras/Model#load_weights>`__.
        """

        self.model.load_weights(filepath, **kwargs)


    def save_weights(self, filepath, **kwargs):
        """
        Save model weights to a HDF5 or TensorFlow file.

        Args:
            filepath (str): filepath to save model weights.  Using a file suffix of
                '.h5' or '.keras' will save in HDF5 format. Otherwise will save as
                TensorFlow.
            kwargs: optional keyword arguments passed to tf save_weights
                method,  see `Tensflow Docs <https://www.tensorflow.org/api_docs/python/tf/keras/Model#save_weights>`__.
        """

        path = os.path.abspath(filepath)
        dir = os.path.dirname(path)
        if not (os.path.exists(dir) and os.path.isdir(dir)):
            os.mkdir(dir)

        self.model.save_weights(path, **kwargs)


    def save_model(self, filepath, **kwargs):
        """
        Save entire model

        Args:
            filepath (str): path to SavedModel or H5 file to save the model.
            kwargs: optional keyword arguments pass to tf save method,
                see `TensorFlow Docs <https://www.tensorflow.org/api_docs/python/tf/keras/Model#save>`__.
        """

        path = os.path.abspath(filepath)
        dir = os.path.dirname(path)
        if not (os.path.exists(dir) and os.path.isdir(dir)):
            os.mkdir(dir)

        self.model.save(path, **kwargs)


    @staticmethod
    def load(filepath, **kwargs):
        """
        Load entire model and return an istantiated TFEstimatorBase class with
        the saved model loaded into it.

        Args:
            filepath (str): path to SavedModel or H5 file of saved model.
            kwargs: Keyword arguments to pass into load_model function. See
                `TensorFlow Doc <`https://www.tensorflow.org/api_docs/python/tf/keras/models/load_model>`__.
        """

        # Add spiegelib custom objects
        custom_objects = {'rms_error': TFEstimatorBase.rms_error,
                          'HighwayLayer': HighwayLayer}

        if 'custom_objects' in kwargs:
            custom_objects.update(kwargs['custom_objects'])
            del kwargs['custom_objects']


        model = GenericTFModel()
        model.model = tf.keras.models.load_model(filepath, custom_objects=custom_objects)
        model.input_shape = model.model.get_layer(index=0).input_shape[1:]
        return model


    @staticmethod
    def rms_error(y_true, y_pred):
        """
        Static method for calculating root mean squared error between predictions
        and targets

        Args:
            y_true (Tensor): Ground truth labels
            y_pred (Tensor): Predictions
        """

        return tf.sqrt(tf.reduce_mean(tf.square(tf.subtract(y_true, y_pred))))


class GenericTFModel(TFEstimatorBase):
    """
    A generic and empty implentation TFEstimatorBase for loading models into
    """

    def build_model(self):
        """
        Empty Model
        """
        pass



#!/usr/bin/env python
"""
Abstract Base Class for Audio Features
"""

from abc import ABC, abstractmethod
import functools
import os.path
import numpy as np
import joblib
from tqdm import trange

from spiegelib.features import StandardScaler


class FeaturesBase(ABC):
    """
    :param sample_rate: Audio sample rate, defaults to 44100
    :type sample_rate: int, optional
    :param frame_size: frame size in audio samples, defaults to 2048
    :type frame_size: int, optional
    :param hop_size: hop size in audio samples, defaults to 512
    :type hop_size: int, optional
    :param time_major: indicates orientation of matrix that features are returned in,
        time_major is (time_slices, features). Defaults to (features, time_slices).
    :type time_major: boolean, optional
    :param per_feature_normalize: Whether normalization should be applied to each
        dimension independently, or on entire feature space. Defaults to True, which
        applies normalization for each dimension independently.
    :type per_feature_normalize: bool, optional
    """

    def __init__(self, sample_rate=44100, frame_size=2048, hop_size=512,
                 time_major=False, scale=False, scale_axis=(0,)):
        """
        Constructor
        """

        super().__init__()

        self.sample_rate = sample_rate
        self.frame_size = frame_size
        self.hop_size = hop_size
        self.time_major = time_major

        self.should_scale = scale
        self.scaler = None
        self.scale_axis = scale_axis
        self.ScalerClass = StandardScaler

        self.input_modifiers = []
        self.prenorm_modifiers = []
        self.output_modifiers = []

        # Update this in inheriting classes if you need
        self.dtype = np.float32


    def __call__(self, audio, scale=None):
        """
        Calls the get_features method. Applies data modifiers prior to and after
        feature extraction

        :param audio: Audio to process features on
        :type audio: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        :param normalize: If set, will override the normalize attribute set
            during construction.
        :type normalize: bool, optional
        :returns: results from audio feature extraction
        :rtype: np.array
        """

        # Input data modification
        for modifer in self.input_modifiers:
            audio = modifer(audio)

        # Run feature extraction
        features = self.get_features(audio)

        # Apply any prenormalization data modification
        for modifier in self.prenorm_modifiers:
            features = modifier(features)

        # Normalize features
        should_scale = scale if scale != None else self.should_scale
        if should_scale:
            assert self.has_scaler(), "Scaler must be set first."
            features = self.scale(features)

        # Apply any output data modification
        for modifier in self.output_modifiers:
            features = modifier(features)

        return features


    def add_modifier(self, modifier, type):
        """
        Add a data modifier to the feature extraction pipeline

        :param modifier: data modifier function. Should accept an np.array,
            apply some modification to that, and return a np.array
        :type modifier: lambda
        :param type: Where to add this into the pipeline. ('input', 'prenormalize', or 'output')
        :type type: str
        """

        if type == 'input':
            self.input_modifiers.append(modifier)
        elif type == 'prenormalize':
            self.prenorm_modifiers.append(modifier)
        elif type == 'output':
            self.output_modifiers.append(modifier)
        else:
            raise ValueError('Type must be one of ("input", "prenormalize", or '
                             '"output"), received %s' % type)


    @abstractmethod
    def get_features(self, audio):
        """
        Must be implemented. Run audio feature extraction on audio provided as parameter.
        Normalization should be applied based on the normalize parameter.

        :param audio: Audio to process features on
        :type audio: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        :returns: results from audio feature extraction
        :rtype: np.array
        """
        pass


    def set_scaler(self, scaler):
        """
        Set a scaler for a dimension, this will be used to normalize that dimension

        :param scaler: A trained scaler object
        :type scaler:
        """

        self.scaler = scaler


    def fit_scaler(self, data, transform=True):
        """
        Fit scaler to dataset for future transforms.

        :param data: data to train scaler on
        :type data: np.ndarray
        :returns: np.ndarray with scaled data
        :rtype: np.ndarray
        """

        # Fit a new scaler to dataset
        self.scaler = self.ScalerClass()
        self.scaler.fit(data, self.scale_axis)

        # Scale the input dataset
        if transform:
            return self.scaler.transform(data)
        else:
            return None


    def scale(self, data):
        """
        Scale features using pre-trained scaler

        :param data: data to be normalized
        :type data: np.array
        :returns: scaled data
        :rtype: np.array
        """

        assert self.has_scaler(), "Scaler must be set first"
        return self.scaler.transform(data)


    def has_scaler(self):
        """
        :returns: a boolean indicating whether or not scaler has been set
        :rtype: boolean
        """
        return self.scaler != None


    def save_scaler(self, location):
        """
        Save the trained scaler for these features for later use

        :param location: Location to save pickled scaler
        :type location: str
        """

        joblib.dump(self.scaler, location)


    def load_scaler(self, location):
        """
        Load trained scaler from disk

        :param location: Pickled file of trained scaler
        :type location: str
        """

        self.scaler = joblib.load(location)



#!/usr/bin/env python
"""
Simple Multi-layer Perceptron Deep Learning Model

Based on model proposed by Yee-King et al. [1]_
"""

import tensorflow as tf
from tensorflow.keras import layers

from spiegelib.estimator.tf_estimator_base import TFEstimatorBase

class MLP(TFEstimatorBase):
    """
    :param input_shape: Shape of matrix that will be passed to model input
    :type input_shape: tuple
    :param num_outputs: Number of outputs the model has
    :type numOuputs: int
    :param kwargs: optional keyword arguments to pass to :class:`spiegelib.estimator.TFEstimatorBase`
    """

    def __init__(self, input_shape, num_outputs, **kwargs):
        """
        Constructor
        """

        super().__init__(input_shape, num_outputs, **kwargs)


    def build_model(self):
        """
        Construct MLP Model
        """

        self.model = tf.keras.Sequential()
        self.model.add(layers.Dense(50, input_shape=self.input_shape, activation='relu'))
        self.model.add(layers.Dense(40, activation='relu'))
        self.model.add(layers.Dense(30, activation='relu'))
        self.model.add(layers.Dense(self.num_outputs))

        self.model.compile(
            optimizer=tf.optimizers.Adam(),
            loss=TFEstimatorBase.rms_error,
            metrics=['accuracy']
        )



#!/usr/bin/env python
"""
STFT Audio Feature Extractor
"""

import numpy as np
import librosa
from spiegelib import AudioBuffer
from spiegelib.features.features_base import FeaturesBase
import spiegelib.core.utils as utils

class STFT(FeaturesBase):
    """
    :param fft_size: number of FFT bins, defaults to 1024 (overrides frame_size)
    :type fft_size: int, optional
    :param kwargs: keyword arguments for base class, see :class:`spiegelib.features.features_base.FeaturesBase`.
    """

    def __init__(self, fft_size=1024, output='complex', scale_axis=None, **kwargs):
        """
        Contructor
        """

        super().__init__(scale_axis=scale_axis, **kwargs)

        if output not in utils.spectrum_types:
            raise TypeError('output must be one of %s' % utils.spectrum_types)

        self.output = output
        self.frame_size = fft_size
        self.dtype = np.float32
        self.complex_dtype = np.complex64


    def get_features(self, audio):
        """
        Run audio feature extraction on audio provided as parameter.
        Normalization should be applied based on the normalize parameter.

        :param audio: Audio to process features on
        :type audio: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        :param normalize: Whether or not the features are normalized, defaults to False
        :type normalize: bool, optional
        :returns: results from audio feature extraction
        :rtype: np.array
        """

        if not isinstance(audio, AudioBuffer):
            raise TypeError('audio must be AudioBuffer, recieved %s' % type(audio))

        features = librosa.stft(
            y=audio.get_audio(),
            n_fft=self.frame_size,
            hop_length=self.hop_size,
        )

        features = utils.convert_spectrum(features, self.output, dtype=self.dtype,
                                          complex_dtype=self.complex_dtype)

        if self.time_major:
            features = np.swapaxes(features, 0, 1)

        return features



#!/usr/bin/env python
"""
LSTM Deep Learning Model

Based on model proposed by Yee-King et al. [1]_
"""

import tensorflow as tf
from tensorflow.keras import layers

from spiegelib.estimator.tf_estimator_base import TFEstimatorBase

class LSTM(TFEstimatorBase):
    """
    :param input_shape: Shape of matrix that will be passed to model input
    :type input_shape: tuple
    :param num_outputs: Number of outputs the model has
    :type numOuputs: int
    :param kwargs: optional keyword arguments to pass to
        :class:`spiegelib.estimator.TFEstimatorBase`
    """

    def __init__(self, input_shape, num_outputs, **kwargs):
        """
        Constructor
        """

        super().__init__(input_shape, num_outputs, **kwargs)


    def build_model(self, highway_layers=100):
        """
        Construct LSTM Model

        :param highway_layers: dimensionality of outer space of hidden layers,
            defaults to 100
        :type highway_layers: int, optional
        """

        self.model = tf.keras.Sequential()
        self.model.add(layers.LSTM(highway_layers, input_shape=self.input_shape,
                       return_sequences=True))
        self.model.add(layers.LSTM(highway_layers, return_sequences=True))
        self.model.add(layers.LSTM(highway_layers))
        self.model.add(layers.Dropout(0.2))
        self.model.add(layers.Dense(
            self.num_outputs,
            use_bias=True,
            kernel_initializer=tf.random_normal_initializer(stddev=0.01),
            bias_initializer=tf.random_normal_initializer(stddev=0.01),
        ))

        self.model.compile(
            optimizer=tf.optimizers.Adam(),
            loss=TFEstimatorBase.rms_error,
            metrics=['accuracy']
        )



#!/usr/bin/python
"""
Basic single objective genetic algorithm
"""

import numpy as np
import random

from deap import algorithms
from deap import base
from deap import creator
from deap import tools

from spiegelib.evaluation.evaluation_base import EvaluationBase
from spiegelib.estimator.estimator_base import EstimatorBase
from spiegelib.synth.synth_base import SynthBase
from spiegelib.features.features_base import FeaturesBase


class BasicGA(EstimatorBase):
    """
    Args:
        synth (Object): Instance of :class:`spiegelib.synth.SynthBase`
        features (Object): Instance of :class:`spiegelib.features.FeaturesBase`.
            This is used in the evaluation function to determine *fitness* of an individual.
        seed (int, optional): Seed for random. Defaults to current system time.
        pop_size (int, optional): Size of population at each generation
        ngen (int, optional): Number of generations to run
        cxpb (float, optional): Crossover probability, must be between 0 and 1.
        mutpb (float, optional): Mutation probability, must be between 0 and 1.
    """

    def __init__(self, synth, features, seed=None, pop_size=100, ngen=25,
                 cxpb=0.5, mutpb=0.3):
        """
        Constructor
        """
        super().__init__()

        if not isinstance(synth, SynthBase):
            raise TypeError("synth must be of type SynthBase")

        self.synth = synth
        self.num_params = len(synth.get_patch())

        if not isinstance(features, FeaturesBase):
            raise TypeError("features must be of type FeaturesBase")

        self.features = features
        self.target = None

        self.pop_size = pop_size
        self.ngen = ngen
        self.cxpb = cxpb
        self.mutpb = mutpb

        self.logbook = None

        random.seed(seed)
        self._setup()


    def _setup(self):
        """
        Setup genetic algorithm
        """

        # Create individual types
        creator.create("BasicGAFitnessMin", base.Fitness, weights=(-1.0,))
        creator.create("BasicGAIndividual", list, fitness=creator.BasicGAFitnessMin)

        # Setup toolbox
        self.toolbox = base.Toolbox()

        # Attribute generator
        self.toolbox.register("attr_float", random.random)

        # Structure initializers
        self.toolbox.register("individual", tools.initRepeat, creator.BasicGAIndividual,
                              self.toolbox.attr_float, self.num_params)

        self.toolbox.register("population", tools.initRepeat, list,
                              self.toolbox.individual)

        self.toolbox.register("evaluate", self.fitness)
        self.toolbox.register("mate", tools.cxTwoPoint)
        self.toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
        self.toolbox.register("select", tools.selTournament, tournsize=3)


    def fitness(self, individual):
        """
        This is automatically called during prediction. Evaluation that calculates
        the fitness of an individual. The individual is
        a new estimated synthesizer parameter setting, and fitness is calculated
        by rendering an audio sample using those parameter settings and then measuring
        the error between that sample and the target sound using audio feature extraction.
        The type of feature extraction is set in the constructor.

        Args:
            individual (list): List of float values representing a synthesizer patch

        Returns:
            list: A list of the error values (b/c this is a single objective GA \
                the list only has one element)
        """

        self.synth.set_patch(individual)
        self.synth.render_patch()
        out = self.synth.get_audio()
        out_features = self.features(out)
        error = EvaluationBase.mean_abs_error(self.target, out_features)
        return error,


    def predict(self, input):
        """
        Run GA prection on input audio target

        Args:
            input (:ref:`AudioBuffer <audio_buffer>`): AudioBuffer to use as target
        """


        self.target = self.features(input)

        pop = self.toolbox.population(n=self.pop_size)
        hof = tools.HallOfFame(1)
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean)
        stats.register("std", np.std)
        stats.register("min", np.min)
        stats.register("max", np.max)

        pop, self.logbook = algorithms.eaSimple(pop, self.toolbox, cxpb=self.cxpb,
                                                mutpb=self.mutpb, ngen=self.ngen,
                                                stats=stats, halloffame=hof,
                                                verbose=True)

        return hof[0]



#!/usr/bin/env python

"""
:class:`SynthVST` is a class for interacting with VST synthesizer plugins.

This class relies on the librenderman package developed by Leon Feddon for interacting
with VSTs. A VST can be loaded, parameters displayed and modified, random patches generated,
and audio rendered for further processing.
"""

from __future__ import print_function
import numpy as np
import librenderman as rm

from spiegelib import AudioBuffer
from spiegelib.synth.synth_base import SynthBase


class SynthVST(SynthBase):
    """
    :param plugin_path: path to vst plugin binary, defaults to None
    :type plugin_path: str, optional
    :param keyword arguments: see :class:`spiegelib.synth.synth_base.SynthBase` for details
    """

    def __init__(self, plugin_path=None, **kwargs):
        super().__init__(**kwargs)

        if plugin_path:
            self.load_plugin(plugin_path)

        else:
            self.engine = None
            self.loaded_plugin = False


    def load_plugin(self, plugin_path):
        """
        Loads a synth VST plugin

        :param plugin_path: path to vst plugin binary
        :type plugin_path: str
        """
        try:
            self.engine = rm.RenderEngine(self.sample_rate, self.buffer_size, self.buffer_size)

            if self.engine.load_plugin(plugin_path):
                self.loaded_plugin = True
                self.generator = rm.PatchGenerator(self.engine)
                self.patch = self.engine.get_patch()
                self.parameters = parse_parameters(self.engine.get_plugin_parameters_description())

                for i in range(len(self.overridden_params)):
                    index, value = self.overridden_params[i]
                    self.engine.override_plugin_parameter(int(index), value)

            else:
                raise Exception('Could not load VST at path: %s' % plugin_path)

        except Exception as error:
            print(error)


    def load_patch(self):
        """
        Update patch parameter. Overridden parameters will not be effected.
        """

        # Check for parameters to include in patch update
        for param in self.patch:
            if not self.is_valid_parameter_setting(param):
                raise Exception(
                    'Parameter %s is invalid. Must be a valid '
                    'parameter number and be in range 0-1. '
                    'Received %s' % param
                )

        # Patch VST with parameters
        self.engine.set_patch(self.patch)


    def is_valid_parameter_setting(self, parameter):
        """
        Checks to see if a parameter is valid for the currently loaded synth.

        :param parameter: A parameter tuple with form `(parameter_index, parameter_value)`
        :type parameter: tuple
        """
        return (
            parameter[0] in self.parameters
            and parameter[1] >= 0.0
            and parameter[1] <= 1.0
        )



    def render_patch(self):
        """
        Render the current patch. Uses the values of midi_note, midi_velocity, note_length_secs,
        and render_length_secs to render audio. Plugin must be loaded first.
        """
        if self.loaded_plugin:
            self.engine.render_patch(
                self.midi_note,
                self.midi_velocity,
                self.note_length_secs,
                self.render_length_secs
            )
            self.rendered_patch = True

        else:
            print("Please load plugin first.")


    def get_audio(self):
        """
        Return monophonic audio from rendered patch

        :return: An audio buffer of the rendered patch
        :rtype: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        """

        if self.rendered_patch:
            audio = AudioBuffer(self.engine.get_audio_frames(), self.sample_rate)
            return audio

        else:
            raise Exception('Patch must be rendered before audio can be retrieved')


    def randomize_patch(self):
        """
        Randomize the current patch. Overridden parameteres will be unaffected.
        """

        if self.loaded_plugin:
            random_patch = self.generator.get_random_patch()
            self.set_patch(random_patch)

        else:
            print("Please load plugin first.")



################################################################################


def parse_parameters(param_str):
    """
    Parse parameter string return by librenderman into a dictionary keyed on parameter
    index with values being the name / short descriptions for the parameter at that index.

    :param param_str: A parameter decription string returned by librenderman
    :type param_str: str
    :returns: A dictionary with parameter index as keys and parameter name / description for values
    :rtype: dict
    """

    param_list = param_str.split("\n")
    param_dict = {}

    for param in param_list:
        param_vals = param.split(":")
        try:
            param_index = int(param_vals[0])
            param_name = param_vals[1].strip()
            param_dict[param_index] = param_name
        except ValueError:
            continue

    return param_dict



#!/usr/bin/env python
"""
Class for storing and passing audio signals around. Can be instantiated with an
array of audio samples or location to load audio from disk. Stores sample rate,
number of channels, and the file name if loaded from disk.

Also has utility functions for loading and saving audio from disk, normalizing,
resizing buffers, and plotting spectrograms.

Examples
^^^^^^^^

Create a new empty ``AudioBuffer``::

    >>> audio = spiegelib.AudioBuffer()

Create an ``AudioBuffer`` from existing audio samples::

    >>> audio = spiegelib.AudioBuffer(audio_samples, 44100)

Load audio from disk into a new ``AudioBuffer``::

    >>> audio = spiegelib.AudioBuffer('./some_audio_file.wav')

Load an entire folder of audio samples into a list of ``AudioBuffer`` objects::

    >>> audio_files = spiegelib.AudioBuffer.load_folder('./audio_folder')

Plot a spectrogram

.. code-block:: python
    :linenos:

    import spiegelib as spgl
    import matplotlib.pyplot as plt

    audio = spgl.AudioBuffer('./audio_file.wav')
    audio.plot_spectrogram()
    plt.show()

.. image:: ./images/audio_buffer_spect.png

"""

import os
import numbers

import numpy as np
import librosa
import librosa.display
import scipy.io.wavfile

import spiegelib.core.utils as utils


class AudioBuffer():
    """
    Args:
        input (np.ndarray, list, str, file-like object): Can be an array
            of audio samples (np.ndarray or list), or a path to a location of an
            audio file on disk. Defaults to None.
        sample_rate (int): Rate of sampled audio if audio data was passed in, or rate
            to resample audio data loaded from disk at, defaults to None. If loading audio
            from a file, will automatically resample to 44100 if no sample rate provided.
        kwargs: keyword args to pass into `librosa load function <https://librosa.github.io/librosa/generated/librosa.core.load.html>`_.

    """

    def __init__(self, input=None, sample_rate=None, **kwargs):
        """
        Constructor
        """

        self.audio = None
        self.sample_rate = None
        self.channels = 0

        # Will be set if audio is loaded from a file
        self.file_name = ''

        path = None
        audio = None

        # If the input is a numpy array or list with numbers then assume
        # we've been passed audio data. Otherwise, assume we've been passed
        # a path to an audio file
        if isinstance(input, np.ndarray):
            audio = input
        elif isinstance(input, list):
            audio = np.array(input)
        elif input:
            path = input

        audio_set = isinstance(audio, np.ndarray)

        # If an audio file location was provided, load audio from disk on construction
        if path:
            self.load(path, sample_rate=(sample_rate if sample_rate else 44100), **kwargs)

        # If audio data was passed in directly
        elif audio_set and sample_rate:
            self.replace_audio_data(audio, sample_rate)

        # raise exception if auio data passed in but not sample_rate
        elif audio_set and not sample_rate:
            raise Exception('Sample rate is required when initializing with audio data')


    def get_audio(self):
        """
        Getter for audio data

        Returns:
            np.ndarray: Array of audio samples
        """
        return self.audio


    def get_sample_rate(self):
        """
        Getter for sample rate

        Returns:
            int: Sample rate of audio buffer
        """
        return self.sample_rate


    def load(self, path, sample_rate=44100, **kwargs):
        """
        Read audio from a file into a numpy array at specific audio rate. Stores
        in this ``AudioBuffer`` object.

        Args:
            path (str, int, pathlib.Path, or file-like object): location of audio file on disk
            sample_rate (int): resample audio to this rate, defaults to 44100. None uses
                native sampling rate.
            kwargs: keyword args to pass into `librosa load function <https://librosa.github.io/librosa/generated/librosa.core.load.html>`_.
        """

        self.audio, self.sample_rate = librosa.core.load(path, sr=sample_rate, **kwargs)
        self.channels = len(self.audio.shape)
        self.file_name = path


    def plot_spectrogram(self, **kwargs):
        """
        Plot spectrogram of this audio buffer. Uses librosas `specshow <https://librosa.github.io/librosa/generated/librosa.display.specshow.html>`_
        function. Uses matplotlib.pyplot to plot spectrogram.

        Defaults to logarithmic y-axis in Hertz and seconds along the x-axis.

        Args:
            kwargs: see `specshow <https://librosa.github.io/librosa/generated/librosa.display.specshow.html>`_.

        Returns:
            axes: The axis handle for figures
        """

        D = librosa.amplitude_to_db(np.abs(librosa.stft(self.get_audio())), ref=np.max)
        y_axis = 'log' if not 'y_axis' in kwargs else kwargs['y_axis']
        x_axis = 's' if not 'y_axis' in kwargs else kwargs['x_axis']
        return librosa.display.specshow(D, y_axis=y_axis, x_axis=x_axis, **kwargs)


    def replace_audio_data(self, audio, sample_rate):
        """
        Replace audio data in this object

        Args:
            audio (np.ndarray): Array of audio samples
            sample_rate (int): Rate that audio data was sampled at
        """

        self.audio = audio
        self.channels = len(self.audio.shape)
        self.sample_rate = sample_rate


    def resize(self, num_samples, start=0):
        """
        Resize audio to a set number of samples. If new length is less than the
        current audio buffer size, then the buffer will be trimmed. If the new
        length is greater than current audio buffer, then the resulting buffer
        will be zero-padded at the end.

        Args:
            num_samples (int): New length of audio buffer in samples
            start_sample (int): Start reading from certain number of samples into
                current buffer
        """

        new_shape = (self.channels, num_samples) if self.channels > 1 else (num_samples,)
        new_audio = np.zeros(new_shape)
        current = self.audio.shape[1] if self.channels > 1 else self.audio.shape[0]
        smaller = min(current-start, num_samples)

        if self.channels > 1:
            new_audio[:,0:smaller] = self.audio[:,start:start+smaller]
        else:
            new_audio[0:smaller] = self.audio[start:start+smaller]

        self.audio = new_audio


    def save(self, path, normalize=False):
        """
        Save an audio to disk as WAV file at given sample rate. Can normalize
        before saving as well.

        Args:
            path (str or open file handle): location to save audio to
            normalize (boolean): normalize audio before writing, defaults to false
        """

        # Make directory if it doesn't exist
        fullpath = os.path.abspath(path)
        dir = os.path.dirname(fullpath)
        if not os.path.exists(dir):
            os.makedirs(dir)

        audio = np.copy(self.audio)
        if normalize:
            audio = AudioBuffer.peak_normalize(audio)

        if self.channels > 1:
            audio = np.transpose(audio)

        scipy.io.wavfile.write(
            fullpath,
            self.sample_rate,
            audio
        )


    @staticmethod
    def peak_normalize(audio):
        """
        Peak normalize audio data

        Args:
            audio (np.ndarray) : array of audio samples

        Returns:
            np.ndarray : normalized array of audio samples
        """

        maxSample = np.max(np.abs(audio))
        minSample = np.min(np.abs(audio))

        if maxSample > minSample:
            audio /= maxSample

        return audio


    @staticmethod
    def load_folder(path, sort=True, sample_rate=44100, **kwargs):
        """
        Try to load a folder of audio samples

        Args:
            path (str): Path to directory of audio files
            sort (bool): Apply natural sort to file names. Default True
            sample_rate (int): Sample rate to load audio files at. Defualts to 44100
            kwargs: keyword args to pass into `librosa load function <https://librosa.github.io/librosa/generated/librosa.core.load.html>`_.

        Returns:
            list: List of ``AudioBuffer`` objects
        """

        abspath = os.path.abspath(path)
        if not (os.path.exists(abspath) and os.path.isdir(abspath)):
            raise ValueError('%s is not a directory' % path)

        dir = [file for file in os.listdir(abspath) if not file.startswith('.')]
        if sort:
            dir.sort(key=utils.natural_keys)

        audio_files = []
        for file in dir:
            try:
                audioFile = AudioBuffer(os.path.join(abspath, file), sample_rate, **kwargs)
                audioFile.file_name = file
                audio_files.append(audioFile)
            except:
                pass

        return audio_files



#!/usr/bin/env python
"""
This class can be used to generate datasets from instances of synthesizers. This is
useful for creating large sets of data for training and validating deep learning models.

Examples
^^^^^^^^

Example generating 50000 training samples and 10000 testing samples from the
*Dexed* VST FM Synthesizer. Each sample is created by creating a random
patch configuration in *Dexed*, and then rendering a one second audio clip of
that patch. A 13-band MFCC is computed on the resulting audio. These audio features
and the synthesizer parameters used to synthesize the audio are saved in numpy files.
Audio features are standardized by removing the mean and scaling to unit variance. The
values used for scaling are saved after the first dataset generation so they
can be used on future data.

.. code-block:: python
    :linenos:

    import spiegelib as spgl

    synth = spgl.synth.SynthVST("/Library/Audio/Plug-Ins/VST/Dexed.vst",
                                note_length_secs=1.0,
                                render_length_secs=1.0)

    # Mel-frequency Cepstral Coefficients audio feature extractor.
    features = spgl.features.MFCC(num_mfccs=13, frame_size=2048,
                                  hop_size=1024 time_major=True)

    # Setup generator for MFCC output and generate 50000 training examples
    # and 10000 testing examples
    generator = spgl.DatasetGenerator(synth, features,
                                      output_folder="./data_FM_mfcc",
                                      normalize=True)
    generator.generate(50000, file_prefix="train_")
    generator.generate(10000, file_prefix="test_")
    generator.save_scaler('data_scaler.pkl')

"""

import os

import numpy as np
import scipy.io.wavfile
from tqdm import trange

from spiegelib.features.features_base import FeaturesBase
from spiegelib.synth.synth_base import SynthBase


class DatasetGenerator():
    """

    Args:
        synth (Object): Synthesizer to generate test data from. Must inherit from
            :class:`spiegelib.synth.SynthBase`.
        features (Object): Features to use for dataset generation. Must inherit from
            :class:`spiegelib.features.FeaturesBase`.
        output_folder (str, optional): Output folder for dataset, defaults to currect working directory.
        save_audio (bool, optional): Whether or not to save rendered audio files, defaults to False.
        scale (bool, optional): Whether or not to scale resulting feature vector. If the feature
            object does not have a scaler set, then this will train a data scaler based on the
            generated dataset and store them in the features object. Call :py:meth:`save_scaler`
            to store scaler settings. Defaults to False.

    Attributes:
        features_filename (str): filename for features output file, defaults to features.npy
        patches_filename (str): filename for patches output file, defaults to patches.npy
        audio_folder_name (str): folder name for the audio output if used. Will be automatically
            created within the output folder if saving audio. Defaults to audio
    """

    def __init__(self, synth, features, output_folder=os.getcwd(), save_audio=False, scale=False):
        """
        Contructor
        """

        # Check for valid synth
        if isinstance(synth, SynthBase):
            self.synth = synth
        else:
            raise TypeError('synth must inherit from SynthBase')

        # Check for valid features
        if isinstance(features, FeaturesBase):
            self.features = features
        else:
            raise TypeError('features must inherit from FeaturesBase')

        # Check for valid output folder
        self.output_folder = os.path.abspath(output_folder)
        if not (os.path.exists(self.output_folder) and os.path.isdir(self.output_folder)):
            os.mkdir(self.output_folder)

        self.save_audio = save_audio

        # Default folder for audio output
        self.audio_folder_name = "audio"

        # Default filenames for output files
        self.features_filename = "features.npy"
        self.patches_filename = "patches.npy"

        # Should the feature set data be scaled?
        self.should_scale = scale


    def generate(self, size, file_prefix="", fit_scaler_only=False):
        """
        Generate dataset with a set of random patches. Saves the extracted features
        and parameter settings in separate .npy files. Files are stored in the output
        folder set during construction (defaults to current working directory) and
        saves the features as "features.npy" and patches as "patches.npy". These file
        names can be prefixed with a string set by the file_prefix argument. If audio
        files are being saved (configured during construction), then the audio files
        are saved in a separate audio folder and all audio files are also prefixed
        by the file_prefix.

        Args:
            size (int): Number of different synthesizer patches to render.
            file_prefix (str, optional): filename prefix for all output data.
            fit_scaler_only (bool, optional): If this is set to True, then
                no data will be saved and only scaler will be set or reset
                for the feature object.
        """

        # Get a single example to determine required array size required
        audio = self.synth.get_random_example()
        features = self.features(audio)
        patch = self.synth.get_patch()

        # Arrays to hold dataset
        shape = list(features.shape)
        shape.insert(0, size)
        feature_set = np.empty(shape, dtype=features.dtype)
        patch_set = np.zeros((size, len(patch)), dtype=np.float32)

        # Should the features be normalized with the feature scaler?
        should_scale = self.should_scale and self.features.has_scaler()

        # Generate data
        for i in trange(size, desc="Generating Dataset"):
            audio = self.synth.get_random_example()
            feature_set[i] = self.features(audio, scale=should_scale)
            patch_set[i] = [p[1] for p in self.synth.get_patch()]

            # Save rendered audio if required
            if self.save_audio:
                self._create_audio_folder()
                audio.save(os.path.join(self.audio_folder_path, "%soutput_%s.wav" % (file_prefix, i)))

        # If only fitting scaler, do that and return. Don't save any data
        if fit_scaler_only:
            print("Fitting scaler only", flush=True)
            self.features.fit_scaler(feature_set, transform=False)
            return

        if self.should_scale and not self.features.has_scaler():
            print("Fitting scaler and scaling data", flush=True)
            feature_set = self.features.fit_scaler(feature_set)

        # Save dataset
        np.save(os.path.join(self.output_folder, "%s%s" % (file_prefix, self.features_filename)), feature_set)
        np.save(os.path.join(self.output_folder, "%s%s" % (file_prefix, self.patches_filename)), patch_set)


    def save_scaler(self, file_name):
        """
        Save feature scaler as a pickle file.

        Args:
            file_name (str): file name for scaler pickle file
        """
        self.features.save_scaler(os.path.join(self.output_folder, file_name))


    def _create_audio_folder(self):
        """
        Check for and create the audio output folder if necassary
        """
        self.audio_folder_path = os.path.abspath(os.path.join(self.output_folder, self.audio_folder_name))
        if not (os.path.exists(self.audio_folder_path) and os.path.isdir(self.audio_folder_path)):
            os.mkdir(self.audio_folder_path)



#!/usr/bin/env python
"""
MFCC Audio Feature Extractor
"""

import numpy as np
import librosa
from spiegelib import AudioBuffer
from spiegelib.features.features_base import FeaturesBase

class MFCC(FeaturesBase):
    """
    :param num_mfccs: number of mffcs to return per frame, defaults to 20
    :type num_mfccs: int, optional
    :param kwargs: keyword arguments for base class, see :class:`spiegelib.features.features_base.FeaturesBase`.
    """

    def __init__(self, num_mfccs=20, scale_axis=0, **kwargs):
        """
        Contructor
        """

        self.num_mfccs = num_mfccs
        super().__init__(scale_axis=scale_axis, **kwargs)


    def get_features(self, audio, normalize=False):
        """
        Run audio feature extraction on audio provided as parameter.
        Normalization should be applied based on the normalize parameter.

        :param audio: Audio to process features on
        :type audio: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        :param normalize: Whether or not the features are normalized, defaults to False
        :type normalize: bool, optional
        :returns: results from audio feature extraction
        :rtype: np.array
        """

        if not isinstance(audio, AudioBuffer):
            raise TypeError('audio must be AudioBuffer, recieved %s' % type(audio))

        if audio.get_sample_rate() != self.sample_rate:
            raise ValueError(
                'audio buffer samplerate does not equal feature '
                'extraction rate, %s != %s' % (audio.get_sample_rate(), self.sample_rate)
            )

        features = librosa.feature.mfcc(
            y=audio.get_audio(),
            sr=self.sample_rate,
            n_fft=self.frame_size,
            hop_length=self.hop_size,
            n_mfcc=self.num_mfccs
        )

        if self.time_major:
            features = np.transpose(features)

        return features



#!/usr/bin/env python
"""
Bidirection LSTM with Highway Layers - LSTM++

Based on model proposed by Yee-King et al. [1]_
"""

import tensorflow as tf
from tensorflow.keras import layers

from spiegelib.estimator.highway_layer import HighwayLayer
from spiegelib.estimator.tf_estimator_base import TFEstimatorBase

class HwyBLSTM(TFEstimatorBase):
    """
    :param input_shape: Shape of matrix that will be passed to model input
    :type input_shape: tuple
    :param num_outputs: Number of outputs the model has
    :type num_outputs: int
    :param kwargs: optional keyword arguments to pass to
        :class:`spiegelib.estimator.TFEstimatorBase`
    """

    def __init__(self, input_shape, num_outputs, lstm_size=128,
                 highway_layers=6, **kwargs):
        """
        Constructor
        """

        self.lstm_size = lstm_size
        self.highway_layers = highway_layers
        super().__init__(input_shape, num_outputs, **kwargs)


    def build_model(self):
        """
        Construct LSTM++ Model
        """

        self.model = tf.keras.Sequential()
        self.model.add(
            layers.Bidirectional(
                layers.LSTM(self.lstm_size),
                input_shape=self.input_shape,
                merge_mode='concat'
            )
        )
        self.model.add(layers.Dropout(0.2))
        self.model.add(layers.Dense(
            64,
            activation='elu',
            activity_regularizer=tf.keras.regularizers.l2()
        ))

        self.hwy = None

        # Add highway layers
        for i in range(self.highway_layers):
            self.hwy = HighwayLayer(
                activation='elu',
                transform_dropout=0.2,
                activity_regularizer=tf.keras.regularizers.l2()
            )
            self.model.add(self.hwy)

        self.model.add(layers.Dense(
            self.num_outputs,
            activation='elu',
            use_bias=True,
            kernel_initializer=tf.random_normal_initializer(stddev=0.01),
            bias_initializer=tf.random_normal_initializer(stddev=0.01),
        ))

        self.model.compile(
            optimizer=tf.optimizers.Adam(),
            loss=TFEstimatorBase.rms_error,
            metrics=['accuracy']
        )



#!/usr/bin/env python
"""
Convolutional Neural Network based on the 6-layer deep model proposed by
Barkan et al. [1]_
"""

import tensorflow as tf
from tensorflow.keras import layers

from spiegelib.estimator.tf_estimator_base import TFEstimatorBase

class Conv6(TFEstimatorBase):
    """
    :param input_shape: Shape of matrix that will be passed to model input
    :type input_shape: tuple
    :param num_outputs: Number of outputs the model has
    :type numOuputs: int
    :param kwargs: optional keyword arguments to pass to
        :class:`spiegelib.estimator.TFEstimatorBase`
    """

    def __init__(self, input_shape, num_outputs, **kwargs):
        """
        Constructor
        """

        super().__init__(input_shape, num_outputs, **kwargs)


    def build_model(self):
        """
        Construct 6-layer CNN Model
        """

        self.model = tf.keras.Sequential()
        self.model.add(layers.Conv2D(32, (3, 3), strides=(2,2), dilation_rate=(1,1),
                                     input_shape=self.input_shape,
                                     activation='relu'))
        self.model.add(layers.Conv2D(71, (3, 3), strides=(2,2), dilation_rate=(1,1),
                                     activation='relu'))
        self.model.add(layers.Conv2D(128, (3, 4), strides=(2,3), dilation_rate=(1,1),
                                     activation='relu'))
        self.model.add(layers.Conv2D(128, (3, 3), strides=(2,2), dilation_rate=(1,1),
                                     activation='relu'))
        self.model.add(layers.Conv2D(128, (3, 3), strides=(2,2), dilation_rate=(1,1),
                                     activation='relu'))
        self.model.add(layers.Conv2D(128, (3, 3), strides=(1,2), dilation_rate=(1,1),
                                     activation='relu'))
        self.model.add(layers.Dropout(0.20))
        self.model.add(layers.Flatten())
        self.model.add(layers.Dense(self.num_outputs))

        self.model.compile(
            optimizer=tf.optimizers.Adam(),
            loss=TFEstimatorBase.rms_error,
            metrics=['accuracy']
        )



#!/usr/bin/env python
"""
Synth Abstract Base Class
"""
from __future__ import print_function
import os
import numbers
import json
from abc import ABC, abstractmethod


class SynthBase(ABC):
    """
    :param sample_rate: sampling rate for rendering audio, defaults to 44100
    :type sample_rate: int, optional
    :param buffer_size: buffer size for rendering audio, defaults to 512
    :type buffer_size: int, optional
    :param midi_note: midi note number used for rendering, 0-127, defaults to 40
    :type midi_note: int, optional
    :param midi_velocity: midi velocity used for rendering. 0-127, defaults to 127
    :type midi_velocity: int, optional
    :param note_length_secs: length of midi note in seconds, defaults to 1.0
    :type note_length_secs: float, optional
    :param render_length_secs: length that audio is rendered for in total, defaults to 2.5
    :type render_length_secs: float, optional
    :param overridden_params: a list of tuples containing the parameter index to override and the value to lock that parameter to, defaults to []
    :type overridden_params: list, optional
    :param clamp_params: If true, parameter values will be forced to the acceptable
        range of values. Defaults to True.
    :type clamp_params: bool, optional

    :cvar rendered_patch: indicates whether a patch has been rendered yet, defaults to False
    :vartype rendered_patch: boolean
    :cvar parameters: parameter indices and names
    :vartype parameters: dict
    :cvar patch: current patch values
    :vartype patch: list
    :cvar param_range: Range of acceptable parameter values, defaults to (0.0, 1.0)
    :vartype param_range: tuple
    """

    def __init__(self, sample_rate=44100, buffer_size=512, midi_note=48,
                 midi_velocity=127, note_length_secs=1.0, render_length_secs=2.0,
                 overridden_params=[], clamp_params=True):
        """
        Constructor
        """
        super().__init__()

        self.sample_rate        = sample_rate
        self.buffer_size        = buffer_size
        self.midi_note          = midi_note
        self.midi_velocity      = midi_velocity
        self.note_length_secs   = note_length_secs
        self.render_length_secs = render_length_secs
        self.overridden_params  = overridden_params
        self.clamp_params       = clamp_params

        # Whether or not patch has been rendered by engine
        self.rendered_patch = False

        # Parameters for this synthesizer, must be set
        self.parameters = None

        # Current patch
        self.patch = None

        # Parameter range defaults to [0.0, 1.0]
        self.param_range = (0.0, 1.0)


    def set_patch(self, parameters):
        """
        Set a new patch with updated parameter values with a list of flaots or a
        list of tuples referring to parameter indices and values.

        :param parameters: A list of floats or tuples. If passing in a list of numbers,
            the length must be either the number of non-overridden parameters of the
            number of total parameters for this synth. Parameter values will then be
            mapped to corresponding parameter in order. If a list of tuples is provided,
            the tuples must have the shape (int, float) where the int is the parameter
            number and the float is the parameter value.
        :type parameters: list
        """

        if not len(parameters):
            return

        param_indices = [p for p in self.get_parameters()]
        overridden_indices = [p[0] for p in self.overridden_params]
        non_overridden_indices = list(set(param_indices) - set(overridden_indices))
        new_patch = []

        # If this is just a list of numbers, then try to associate with parameter settings
        if isinstance(parameters[0], numbers.Number):

            # Received same number of parameters as non-overridden parameters,
            # map directly to non-overridden parameters
            if len(parameters) == len(non_overridden_indices):
                new_patch = [(non_overridden_indices[i], float(parameters[i]))
                             for i in range(len(parameters))]

            # Received same number of parameters as total parameter count,
            # map the non-overridden parameters from that list
            elif len(parameters) == len(param_indices):
                    new_patch = [(i, float(parameters[i]))
                                 for i in non_overridden_indices]
            else:
                raise Exception((
                    'Unclear on how to map parameters, received %s parameters '
                    'and there are %s non-overridden parameters and %s total parameters.'
                ) % (len(parameters), len(non_overridden_indices), len(overridden_indices)))

        # If this is a list of tuples then add those directly
        elif len(parameters[0]) == 2:
            new_patch = parameters

        else:
            raise Exception('Invalid parameter list provided. Must be a list of numbers or a list of tuples.')

        # Update patch member variable with new patch, skipping any overridden params
        for param in new_patch:
            if not param[0] in overridden_indices:
                value = param[1]
                if self.clamp_params:
                    value = max(min(value, self.param_range[1]), self.param_range[0])

                self.patch[param[0]] = (param[0], value)

        # Load new patch into synth engine
        self.rendered_patch = False
        self.load_patch()


    @abstractmethod
    def load_patch(self):
        """
        Must be overridden. Load current patch into synth engine
        """
        pass


    @abstractmethod
    def render_patch(self):
        """
        Must be overridden. Should render audio for the currently loaded patch
        """
        pass


    @abstractmethod
    def get_audio(self):
        """
        This method must be overridden and should return an audio buffer rendered
        during the last call to render_patch.

        :return: An audio buffer of float audio samples with a value between -1 & 1
        :rtype: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        """
        pass

    @abstractmethod
    def randomize_patch(self):
        """
        This method must be overridden and should have the effect
        of randomizing parameters of the synthesizer. Overridden methods should be
        uneffected by this randomization
        """
        pass


    def get_random_example(self):
        """
        Returns audio from a new random patch

        :return: An audio buffer
        :rtype: np.array
        """

        self.randomize_patch()
        self.render_patch()
        return self.get_audio()


    def get_parameters(self):
        """
        Returns parameters for the synth

        :return: A dictionary of parameters with the parameter index (int) as the key
            and the parameter name short description as the value
        :rtype: Dictionary
        """

        return self.parameters


    def get_patch(self, skip_overridden=True):
        """
        Get current patch

        :param skip_overridden: Indicates whether to remove overridden parameters
            from results, defaults to True
        :type skip_overridden: bool, optional
        """

        if not skip_overridden:
            return self.patch

        patch = []
        overridden_params = [p[0] for p in self.overridden_params]
        for item in self.patch:
            if not (item[0] in overridden_params):
                patch.append(item)

        return patch


    def set_overridden_parameters(self, parameters):
        """
        Override parameters with specific values

        :param parameters: List of parameter tuples with parameter index and parameter
            value that will be patched and then the parameter frozen.
        :type parameter: list
        """

        overridden_params = []
        for param in parameters:
            overridden_params.append((param[0], param[1]))

        self.overridden_params = []
        self.set_patch(overridden_params)
        self.overridden_params = overridden_params



    def save_state(self, path):
        """
        Save parameters and current state to a JSON file. Includes whether or not
        a parameter has been overridden so this can be used to save synth configuration

        :param path: Location to save JSON file to
        :type path: str
        """

        assert self.parameters, "Parameters must be set before saving parameter state"
        assert self.patch,      "Patch must be set before saving parameter state"

        # Make sure directory exists and create if it doesn't
        fullpath = os.path.abspath(path)
        directory = os.path.dirname(fullpath)

        if not os.path.exists(directory):
            os.mkdir(directory)

        # Create a dictionary of parameters and settings for saving
        param_dict = {}
        overridden_params = [p[0] for p in self.overridden_params]

        for parameter in self.patch:
            param_dict[parameter[0]] = {
                "id": parameter[0],
                "desc": self.parameters[parameter[0]],
                "value": float(parameter[1]),
                "overridden": parameter[0] in overridden_params
            }

        with open(fullpath, 'w') as fp:
            json.dump(param_dict, fp, indent=True, sort_keys=True)


    def load_state(self, path):
        """
        Load parameter state from JSON file. Will set a new patch and overridden
        parameters.

        :param path: Location to load JSON file from
        :type path: str
        """

        fullpath = os.path.abspath(path)
        assert os.path.exists(fullpath), "Path does not exists: %s" % fullpath

        param_dict = {}
        with open(fullpath, 'r') as fp:
            param_dict = json.load(fp)

        patch = []
        overridden = []
        for key in param_dict:
            if param_dict[key]['overridden']:
                overridden.append((
                    param_dict[key]['id'],
                    param_dict[key]['value']
                ))
            else:
                patch.append((
                    param_dict[key]['id'],
                    param_dict[key]['value']
                ))

        self.set_overridden_parameters(overridden)
        self.set_patch(patch)



#!/usr/bin/env python
"""
Data scaler class that implments standarization. This transforms data
to have a mean of zero and variance of 1.
"""
import numpy as np
from spiegelib.features import DataScalerBase

class StandardScaler(DataScalerBase):
    """
    """

    def _reset(self):
        """
        Reset attributes
        """

        if hasattr(self, 'mean'):
            del self.mean
            del self.std


    def fit(self, data, axis=None):
        """
        Compute mean and std for later scaling

        :param data: data to use to calculate mean and std on
        :type data: np.ndarray
        :param axis: axis or axes to use for calculating scaling parameters on.
        :type axis: int, tuple
        """

        self._reset()
        self.mean = data.mean(axis)

        # Calculate standard deviation and handle zeros
        variance = data.var(axis)
        if isinstance(variance, np.ndarray):
            variance[variance == 0.0] = 1.0
            self.std = np.sqrt(variance)

        elif np.isscalar(variance):
            variance = 1.0 if variance == 0.0 else variance
            self.std = np.sqrt(variance)

        self.fit_axis = axis
        if self.fit_axis != None and not isinstance(self.fit_axis, tuple):
            self.fit_axis = (self.fit_axis,)
        self.fit_shape = data.shape


    def transform(self, data):
        """
        Perform normalization on data

        :param data: data to normalize
        :type data: np.array
        :returns: Normalized data
        :rtype: np.ndarray
        """

        if not hasattr(self, 'mean'):
            raise Exception("You must fit this scaler first")

        if self.fit_axis == None:
            return (data - self.mean) / self.std

        mean_expanded = np.copy(self.mean)
        std_expanded = np.copy(self.std)

        if len(data.shape) == len(self.fit_shape):
            for axis in self.fit_axis:
                mean_expanded = np.expand_dims(mean_expanded, axis=axis)
                std_expanded = np.expand_dims(std_expanded, axis=axis)

        elif len(data.shape) == len(self.fit_shape) - 1:
            for axis in self.fit_axis[1:]:
                mean_expanded = np.expand_dims(mean_expanded, axis=(axis-1))
                std_expanded = np.expand_dims(std_expanded, axis=(axis-1))

        else:
            raise ValueError("Input data has incorrect shape. Expected %sD or %D, received %D"
                             % (len(self.fit_shape), len(self.fit_shape) -1, len(data.shape)))

        return (data - mean_expanded) / std_expanded



#!/usr/bin/env python
"""
FFT Audio Feature Extractor
"""

import numpy as np
from spiegelib import AudioBuffer
from spiegelib.features.features_base import FeaturesBase
import spiegelib.core.utils as utils

class FFT(FeaturesBase):
    """
    :param fft_size: Size of FFT, defaults to None. If set, will truncate input
        if smaller than input size, or zero pad if longer.
    :type fft_size: int, optional
    :param kwargs: keyword arguments for base class, see :class:`spiegelib.features.features_base.FeaturesBase`.
    """

    def __init__(self, output='complex', fft_size=None, scale_axis=None, **kwargs):
        """
        Contructor
        """

        # Setup feature base class -- FFT is time summarized, so no
        # time slices are used, defaults to normalizing the entire result
        # as opposed to normalizing across each bin separately.
        super().__init__(scale_axis=scale_axis, **kwargs)

        self.frame_size = fft_size

        if output not in utils.spectrum_types:
            raise TypeError('output must be one of %s' % utils.spectrum_types)

        self.output = output
        self.dtype = np.float32
        self.complex_dtype = np.complex64


    def get_features(self, audio):
        """
        Run audio feature extraction on audio provided as parameter.
        Normalization should be applied based on the normalize parameter.

        :param audio: Audio to process features on
        :type audio: :class:`spiegelib.core.audio_buffer.AudioBuffer`
        :param normalize: Whether or not the features are normalized, defaults to False
        :type normalize: bool, optional
        :returns: results from audio feature extraction
        :rtype: np.array
        """

        if not isinstance(audio, AudioBuffer):
            raise TypeError('audio must be AudioBuffer, recieved %s' % type(audio))

        # Determine the length of audio input and determine number of FFT
        # samples to keep.
        buffer = audio.get_audio()
        n_samples = len(buffer)
        n_output = int((n_samples/2) + 1)
        self.dimensions = n_output

        # Run Fast Fourier Transform
        spectrum = np.fft.fft(audio.get_audio(), n=self.frame_size)[0:n_output]
        features = utils.convert_spectrum(spectrum, self.output, dtype=self.dtype,
                                         complex_dtype=self.complex_dtype)

        return features



#!/usr/bin/env python
"""
Abstract Base Class for objective evaluations
"""

from abc import ABC, abstractmethod
import numpy as np
import matplotlib.pyplot as plt
import json

from spiegelib import AudioBuffer
from spiegelib.core.utils import NumpyNumberEncoder

class EvaluationBase(ABC):
    """
    Example Input::

        targets = [target_1, target_2]
        estimations = [[prediction_1_for_target_1, prediction_2_for_target_1],
                       [prediction_1_for_target_2, prediction_2_for_target_2]]

    Where prediction_1 and prediction_2 would represent results from two different
    methods or sources. For example, audio for all prediction_1 samples might be
    from a GA and all audio for prediction_2 samples might be from a deep learning
    approach.

    Args:
        targets (list): a list of objects to use
            as the ground truth for evaluation
        estimations (lits): a 2D list of objects.
            Should contain a list of objects representing estimations for each target.
            The position of an object in each list is used to distinguish between different
            sources. For example, if you are comparing two different
            synthesizer programming methods, then you would want to make sure to
            have the results from each method in the same position in each list.
    """

    def __init__(self, targets, estimations):
        """
        Constructor
        """

        if not isinstance(targets, list):
            raise TypeError("Expected targets to be a list")

        self.verify_input_list(targets)

        self.num_targets = len(targets)
        self.targets = targets

        for item in estimations:
            if not isinstance(item, list):
                raise TypeError("Expected list of lists for estimations")

            self.verify_input_list(item)

        self.num_sources = len(estimations)
        self.estimations = estimations
        self.scores = {}
        self.stats = {}


    @abstractmethod
    def evaluate_target(self, target, predictions):
        """
        Abstract method. Must be implemented and evaluate a single target and predictions
        made for that target. Called automatically by :func:`~evaluate`

        Args:
            target (list): Audio to use as ground truth in evaluation
            predictions (list): list of :ref:`AudioBuffer <audio_buffer>` objects to evaluate
                against the target audio file.
        Returns:
            list: A list of dictionaries with stats for each prediction evaluation
        """
        pass


    def evaluate(self):
        """
        Run evaluation. Calls evaluate_target on all targets and creates a dictionary
        of metrics stored in the scores attribute.

        Saves each prediction for a target in a dictionary keyed by the position in
        the prediction list that it was constructed with - uses the key 'source_#'.
        Where # is the position index.
        """

        self.stats = {}
        self.scores = {}
        for i in range(len(self.targets)):
            target_estmations = [est[i] for est in self.estimations]
            results = self.evaluate_target(self.targets[i], target_estmations)
            results_dict = {}
            for j in range(len(results)):
                results_dict['source_%s' % j] = results[j]

            self.scores['target_%i' % i] = results_dict


    def get_scores(self):
        """
        Returns:
            dict: scores calculated during evaluation
        """
        if not len(self.scores):
            print(
                "No scores available, did you run evaluate method? If you did "
                "make sure scores are updated in evaluate method."
            )

        return self.scores


    def get_stats(self):
        """
        Returns:
            dict: stats that summarize the scores for each source using mean, median, \
                standard deviation, minimum, and maximum.
        """

        if not len(self.stats):
            self._compute_stats()

        return self.stats


    def save_scores_json(self, path):
        """
        Save scores to a JSON file

        Args:
            path (str): location to save JSON file
        """

        with open(path, 'w') as fp:
            json.dump(self.scores, fp, indent=True, cls=NumpyNumberEncoder)


    def save_stats_json(self, path):
        """
        Save score statistics as a JSON file

        Args:
            path (str): location to save JSON file
        """

        # Try to compute stats if it looks like they haven't been
        if not len(self.stats):
            self._compute_stats()

        # If there are now stats, save those
        if len(self.stats):
            with open(path, 'w') as fp:
                json.dump(self.stats, fp, indent=True, cls=NumpyNumberEncoder)

        else:
            print('No stats to save! Did you run evaluation?')


    def plot_hist(self, sources, metric, bins=None, clip_upper=None, **kwargs):
        """
        Plot a histogram of results of evaluation. Uses Matplotlib.

        Args:
            sources (list): Which audio sources to include in histogram. [0]
                would use the first prediction source passed in during construction,
                [1] would use the seconds, etc.
            metric (str): Which metric to use for creating the histogram. Depends on
                which were used during evaluation.
            bins (int or sequence or str, optional): passed into matplotlib hist
                method and indicates the number of bins to use, or if it is a list
                then it dfines bin edges. With Numpy 1.11 or newer, you can alternatively
                provide a string describing a binning strategy, such as 'auto',
                'sturges', 'fd', 'doane', 'scott', 'rice' or 'sqrt'
            clipper_upper (number, optional): Set an upper range for input values.
                This can be used to force any values above a certain range into the
                right most hitogram bin.
            kwargs: Keyword arguments to be passed into `hist <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.hist.html>`_
        """

        values = []
        for source in sources:
            values.extend([self.scores[key]['source_%s' % source][metric]
                           for key in self.scores])

        values = np.array(values)

        if clip_upper:
            values = np.clip(values, np.min(values), clip_upper)

        plt.hist(np.array(values), bins,
                 alpha=0.9, edgecolor='black')


    def verify_input_list(self, input_list):
        """
        Base method for verifying input list. Override to implement verification.
        For example, override and call :func:`~spiegelib.evaluation.EvaluationBase.verify_audio_input_list`
        on input_list to verify that :ref:`AudioBuffer <audio_buffer>` objects are being
        passed in.
        """
        pass


    @staticmethod
    def verify_audio_input_list(input_list):
        """
        Static method for verifying input lists with audio buffers
        """

        for item in input_list:
            if not isinstance(item, AudioBuffer):
                raise TypeError('Must be an AudioBuffer, received %s' \
                                % type(item))


    @staticmethod
    def euclidian_distance(A, B):
        """
        Calculates the euclidian distance between two arrays.

        Args:
            A (np.ndarray): First array  (Ground truth)
            B (np.ndarray): Second array (Prediction)

        Returns:
            float: Euclidian distance
        """
        return np.linalg.norm(np.subtract(A,B).flatten())


    @staticmethod
    def manhattan_distance(A, B):
        """
        Calculates the manhattan distance between two arrays.

        Args:
            A (np.ndarray): First array  (Ground truth)
            B (np.ndarray): Second array (Prediction)

        Returns:
            float: Manhattan distance
        """
        return np.linalg.norm(np.subtract(A,B).flatten(), 1)


    @staticmethod
    def mean_abs_error(A, B):
        """
        Calculates mean absolute error between two arrays. Mean(ABS(A-B)).

        Args:
            A (np.ndarray): First array  (Ground truth)
            B (np.ndarray): Second array (Prediction)

        Returns:
            float: Mean absolute error
        """
        return np.mean(np.abs(A-B).flatten())


    @staticmethod
    def mean_squared_error(A, B):
        """
        Calculates mean squared error between two arrays. Mean(Square(A-B)).

        Args:
            A (np.ndarray): First array  (Ground truth)
            B (np.ndarray): Second array (Prediction)

        Returns:
            float: Mean squared error
        """
        return np.mean(np.square(A-B).flatten())


    def _scores_to_nparray(self):
        """
        Reorganize scores into a set of np.ndarrays, one for each metric for each
        source. Returned as a dictionary: dict[source][metric]
        """

        summaries = {}
        scores = self.get_scores()
        i = 0

        # Organize scores into arrays organized by the source and the metric
        for target in scores:
            for source in scores[target]:
                if not source in summaries:
                    summaries[source] = {}

                for metric in scores[target][source]:
                    if not metric in summaries[source]:
                        summaries[source][metric] = np.zeros(self.num_targets)

                    summaries[source][metric][i] = scores[target][source][metric]

            i = i + 1

        return summaries


    def _compute_stats(self):
        """
        :returns: A dictionary of scores summarized using mean, median, and stddev
            for each source
        :rtype: dict
        """

        summaries = self._scores_to_nparray()

        # Now compute stats on those lists
        self.stats = {}
        for source in summaries:
            if not source in self.stats:
                self.stats[source] = {}

            for metric in summaries[source]:
                if not metric in self.stats[source]:
                    self.stats[source][metric] = {}

                self.stats[source][metric]['mean'] = np.median(summaries[source][metric])
                self.stats[source][metric]['median'] = np.mean(summaries[source][metric])
                self.stats[source][metric]['std'] = np.std(summaries[source][metric])
                self.stats[source][metric]['min'] = np.min(summaries[source][metric])
                self.stats[source][metric]['max'] = np.max(summaries[source][metric])



#!/usr/bin/env python
"""
Logger class for logging accuracy and loss from epochs when training a
``tf.keras.Model``. Inherits from ``tf.keras.callbacks.Callback`` and should
be passed in as a callback during model training.
"""

import matplotlib.pyplot as plt
from tensorflow.keras.callbacks import Callback

class TFEpochLogger(Callback):
    """
    Attributes:
        log_data (dict): Dictionary of data logged at the end of each epoch
            during training. Keyed on the epoch number.
    """

    def __init__(self):
        """
        Constructor
        """

        self.log_data = {}
        super().__init__()


    def on_epoch_end(self, epoch, logs=None):
        """
        Logging on end of an epoch. This is called automatically during training.
        Stores infomation passed in the logs parameter into ``log_data`` attribute.

        Args:
            epoch (int): current epoch that just finished
            logs: Any data to be logged
        """

        self.log_data[epoch] = logs


    def get_plotting_data(self):
        """
        Helper function to organize log_data for easier plotting
        """

        plot_data = {}
        epochs = []
        for epoch in self.log_data:
            epochs.append(epoch + 1)
            for key in self.log_data[epoch]:
                if key in plot_data:
                    plot_data[key].append(self.log_data[epoch][key])
                else:
                    plot_data[key] = [self.log_data[epoch][key]]

        return epochs, plot_data


    def plot(self):
        """
        Plot logged training and validation data using matplotlib
        """

        epochs, plot_data = self.get_plotting_data()

        train_accuracy = 'accuracy' in plot_data
        train_loss = 'loss' in plot_data
        val_accuracy = 'val_accuracy' in plot_data
        val_loss = 'val_loss' in plot_data

        if train_accuracy and train_loss:
            fig, axs = plt.subplots(2)
            fig.suptitle('Model Accuracy and Loss')

            axs[0].plot(epochs, plot_data['accuracy'], label="Train Accuracy")
            if val_accuracy:
                axs[0].plot(epochs, plot_data['val_accuracy'],
                            label="Validation Accuracy")

            axs[0].set(ylabel='Accuracy (%)')
            axs[0].legend()

            axs[1].plot(epochs, plot_data['loss'], label="Train Loss")
            if val_loss:
                axs[1].plot(epochs, plot_data['val_loss'], label="Validation Loss")

            axs[1].set(xlabel='Epochs', ylabel='Loss')
            axs[1].legend()

            plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Init for synth components
"""

from .synth_base import SynthBase

try:
    import librenderman
    from .synth_vst import SynthVST
except:
    print("librenderman package not installed, SynthVST class is unavailable. To use VSTs please install librenderman.")
    print("https://spiegelib.github.io/spiegelib/getting_started/installation.html")



#!/use/bin/env python
"""
Audio evaluation using MFCC error between targets and estimations. See
:class:`~spiegelib.evaluation.EvaluationBase` for documentation of methods for
running objective evaluation using inheriting classes, such as this one.

Example::

    import spiegelib as spgl

    targets = spgl.AudioBufer('./folder_of_target_audio_files')
    ga_predictions = spgl.AudioBuffer('./folder_of_ga_predictions')
    mlp_predictions = spgl.AudioBuffer('./folder_of_mlp_predictions')
    estimations = [ga_predictions, mlp_predictions]

    # Instantiate evaluation with audio buffers and run evaluation
    evaluation = spgl.evaluation.MFCCEval(targets, estimations)
    evaluation.evaluate()

    # Save raw results and stats of results as JSON
    evaluation.save_scores_json('./mfcc_results.json')
    evaluation.save_stats_json('./mfcc_eval_stats.json')

    # Plot a histogram of GA predictions using mean absolute error metrics
    evaluation.plot_hist([0], 'mean_abs_error')

    # Plot a histogram of MLP predictions using mean absulote error metrics
    evaluation.plot_hist([1], 'mean_abs_error')
"""

from spiegelib.evaluation.evaluation_base import EvaluationBase
from spiegelib.features.mfcc import MFCC


class MFCCEval(EvaluationBase):
    """
    Pass in a list of target :class:`~spiegelib.AudioBuffer` objects and lists of
    estimated :class:`~spiegelib.AudioBuffer` objects for each target.

    Example Input::

        targets = [target_1, target_2]
        estimations = [[prediction_1_for_target_1, prediction_2_for_target_1],
                       [prediction_1_for_target_2, prediction_2_for_target_2]]

    Where prediction_1 and prediction_2 would represent results from two different
    methods or sources. For example, audio for all prediction_1 samples might be
    from a GA and all audio for prediction_2 samples might be from a deep learning
    approach.

    Args:
        targets (list): a list of :ref:`AudioBuffer <audio_buffer>` objects to use
            as the ground truth for evaluation
        estimations (lits): a 2D list of :ref:`AudioBuffer <audio_buffer>` objects.
            Should contain a list of AudioBuffers for each target. The position of an
            AudioBuffer in each list is used to distinguish between different
            audio sources. For example, if you are comparing two different
            synthesizer programming methods, then you would want to make sure to
            have the results from each method in the same position in each list.
        sample_rate (int, optional): sample rate to run feature extraction at. Defaults
            to the audio rate of the first target.
        kwargs: Keyword arguments to pass into :class:`~spiegelib.evaluation.EvaluationBase`
    """

    def __init__(self, targets, estimations, sample_rate=None, **kwargs):
        """
        Constructor
        """
        self.sample_rate = sample_rate if sample_rate else targets[0].get_sample_rate()
        super().__init__(targets, estimations, **kwargs)


    def evaluate_target(self, target, predictions):
        """
        Called automatically by :func:`spiegelib.evaluation.EvaluationBase.evaluate`

        Evaluates difference between a target :ref:`AudioBuffer <audio_buffer>` and
        a list of estimation :ref:`AudioBuffers <audio_buffer>`. Calculates mean absolute
        error, mean squared error, euclidian distance, and manhattan distance on results
        and returns a dictionary for each estimation with results in a list.

        Args:
            target (list): Audio to use as ground truth in evaluation
            predictions (list): list of :ref:`AudioBuffer <audio_buffer>` objects to evaluate
                against the target audio file.
        Returns:
            list: A list of dictionaries with stats for each prediction evaluation
        """

        results = []
        mfcc = MFCC(sample_rate=self.sample_rate)
        target_mfccs = mfcc(target)

        for pred in predictions:
            estimated_mfccs = mfcc(pred)
            results.append({
                'mean_abs_error': EvaluationBase.mean_abs_error(target_mfccs, estimated_mfccs),
                'mean_squared_error': EvaluationBase.mean_squared_error(target_mfccs, estimated_mfccs),
                'euclidian_distance': EvaluationBase.euclidian_distance(target_mfccs, estimated_mfccs),
                'manhattan_distance': EvaluationBase.manhattan_distance(target_mfccs, estimated_mfccs),
            })

        return results


    def verify_input_list(self, input_list):
        """
        Overriding verification method to check for AudioBuffers. Called automatically
        during construction.
        """
        EvaluationBase.verify_audio_input_list(input_list)



#!/usr/bin/env python
"""
Spectral features summarized over time using mean and variance
"""

import numpy as np
import librosa
from spiegelib import AudioBuffer
from spiegelib.features.features_base import FeaturesBase

class SpectralSummarized(FeaturesBase):
    """
    :param kwargs: See :class:`spiegelib.features.features_base.FeaturesBase`
    """

    def __init__(self, scale_axis=0, **kwargs):
        """
        Constructor
        """
        super().__init__(scale_axis=scale_axis, **kwargs)


    def get_features(self, audio, normalize=False):
        """
        Run audio feature extraction on audio provided as parameter.
        Normalization should be applied based on the normalize parameter.

        :param audio: Audio to process features on
        :type audio: np.array
        :param normalize: Whether or not the features are normalized, defaults to False
        :type normalize: bool, optional
        :returns: results from audio feature extraction
        :rtype: np.array
        """

        if not isinstance(audio, AudioBuffer):
            raise TypeError('audio must be AudioBuffer, recieved %s' % type(audio))

        if audio.get_sample_rate() != self.sample_rate:
            raise ValueError(
                'audio buffer samplerate does not equal feature '
                'extraction rate, %s != %s' % (audio.get_sample_rate(), self.sample_rate)
            )

        spectral_centroid = librosa.feature.spectral_centroid(
            y=audio.get_audio(),
            sr=self.sample_rate,
            n_fft=self.frame_size,
            hop_length=self.hop_size,
        )

        spectral_bandwidth = librosa.feature.spectral_bandwidth(
            y=audio.get_audio(),
            sr=self.sample_rate,
            n_fft=self.frame_size,
            hop_length=self.hop_size,
        )

        spectral_contrast = librosa.feature.spectral_contrast(
            y=audio.get_audio(),
            sr=self.sample_rate,
            n_fft=self.frame_size,
            hop_length=self.hop_size,
        )

        spectral_flatness = librosa.feature.spectral_flatness(
            y=audio.get_audio(),
            n_fft=self.frame_size,
            hop_length=self.hop_size,
        )

        spectral_rolloff = librosa.feature.spectral_rolloff(
            y=audio.get_audio(),
            sr=self.sample_rate,
            n_fft=self.frame_size,
            hop_length=self.hop_size,
        )

        features = np.array([
            spectral_centroid.mean(),
            spectral_centroid.var(),
            spectral_bandwidth.mean(),
            spectral_bandwidth.var(),
            spectral_flatness.mean(),
            spectral_flatness.var(),
            spectral_rolloff.mean(),
            spectral_rolloff.var(),
            spectral_contrast[0].mean(),
            spectral_contrast[0].var(),
            spectral_contrast[1].mean(),
            spectral_contrast[1].var(),
            spectral_contrast[2].mean(),
            spectral_contrast[2].var(),
            spectral_contrast[3].mean(),
            spectral_contrast[3].var(),
            spectral_contrast[4].mean(),
            spectral_contrast[4].var(),
            spectral_contrast[5].mean(),
            spectral_contrast[5].var(),
            spectral_contrast[6].mean(),
            spectral_contrast[6].var(),
        ])

        return features



#!/usr/bin/env python
"""
This class performs synthesizer sound matching by estimating parameters for synthesizer
in order to match a target input sound. It accepts an implementation of :ref:`SynthBase <synth_base>`
to estimate parameters for and an implementation of :ref:`EstimatorBase <estimator_base>`, which
performs the parameter estimation. Optionally, feature extraction can be performed on
the target audio file prior to being fed into the estimator using an implementation
of :ref:`FeaturesBase <features_base>`.

Example
^^^^^^^

Sound matching with a genetic algorithm

.. code-block:: python
    :linenos:

    import spiegelib as spgl

    # Load synth
    synth = spgl.synth.SynthVST("/Library/Audio/Plug-Ins/VST/Dexed.vst")

    # Setup a basic GA using MFCC error as the evaluation function
    ga_extractor = spgl.features.MFCC(num_mfccs=13, hop_size=1024)
    ga = spgl.estimator.BasicGA(synth, ga_extractor)


    # Setup sound match object with synth and GA
    ga_matcher = spgl.SoundMatch(synth, ga)

    # Load a target audio file
    target = spgl.AudioBuffer('./target.wav')

    # Perform sound matching on audio target
    result_audio = ga_matcher.match(target)
    result_patch = ga_matcher.get_patch()

"""

import os

import librosa

from spiegelib import AudioBuffer
from spiegelib.synth.synth_base import SynthBase
from spiegelib.features.features_base import FeaturesBase
from spiegelib.estimator.estimator_base import EstimatorBase

class SoundMatch():
    """
    Args:
        synth (Object): must inherit from :class:`spiegelib.synth.SynthBase`
        estimator (Object): must inherit from :class:`spiegelib.estimator.EstimatorBase`
        features (Object, optional): must inherit from :class:`spiegelib.features.FeatureBase`

    Raises:
        TypeError: If synth, estimator, or features parameters do not inherit from
            the correct base class.
    """

    def __init__(self, synth, estimator, features=None):
        """
        Constructor
        """

        # Check for valid synth
        if isinstance(synth, SynthBase):
            self.synth = synth
        else:
            raise TypeError('synth must inherit from SynthBase, received %s' % type(synth))

        # Check for valid estimator
        if isinstance(estimator, EstimatorBase):
            self.estimator = estimator
        else:
            raise TypeError('estimator must inherit from EstimatorBase, received type %s' % type(estimator))

        # Check for valid feature extraction object
        self.features = None
        if features:
            if isinstance(features, FeaturesBase):
                self.features = features
            else:
                raise TypeError('features must inherit from Featurebase, received %s' % type(features))

        self.patch = None


    def get_patch(self):
        """
        Returns:
            dict: The resuling patch after estimation

        Raises:
            Exception: If sound matching has not been run first
        """
        if not self.patch:
            raise Exception('Please run match first')

        return self.patch


    def match(self, target):
        """
        Attempt to estimate parameters for target audio

        Args:
            target (:ref:`AudioBuffer <audio_buffer>`): input audio to use as target

        Returns:
            :ref:`AudioBuffer <audio_buffer>`: audio output from synthesizer after sound matching
        """

        # Attempt to run feature extraction if features have been provided
        if self.features:
            input = self.features(target)
        else:
            input = target

        # Estimate parameters
        params = self.estimator.predict(input)

        # Load patch into synth and return audio
        self.synth.set_patch(params)
        self.synth.render_patch()
        self.patch = self.synth.get_patch()
        return self.synth.get_audio()


    def match_from_file(self, path):
        """
        Load audio file from disk and perform sound matching on it

        Args:
            filepath (str): location of audio file on disk

        Returns:
            :ref:`AudioBuffer <audio_buffer>`: audio output from synthesizer after sound matching
        """

        target = AudioBuffer(path, self.features.sample_rate)
        return self.match(target)



#!/usr/bin/env python

from setuptools import setup
from setuptools import find_packages

__version__ = '0.0.1'
__author__ = "Jordie Shier"
__contact__ = "jordieshier@gmail.com"
__url__ = "https://github.com/spiegelib/spiegelib"
__license__ = "MIT"


with open("README.md", "r", encoding='utf-8') as f:
    readme = f.read()

setup(
    name='spiegelib',
    version=__version__,
    author=__author__,
    author_email=__contact__,
    description='Synthesizer Programming with Intelligent Exploration, Generation, and Evaluation Library',
    long_description=readme,
    long_description_content_type='text/markdown',
    url=__url__,
    licence=__license__,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'librosa',
        'tqdm',
        'matplotlib',
        'tensorflow',
        'deap',
    ],
    extras_require={
        'dev': [
            'sphinx',
            'sphinx_rtd_theme',
            'ipython',
            'pytest',
            'pytest-datadir',
            'pytest-mpl',
            'tox',
            'twine'
        ],
    }
)



#!/usr/bin/env python

"""
Init file for estimators
"""

from .estimator_base import EstimatorBase
from .tf_estimator_base import TFEstimatorBase

# Genetic Algorithms
from .basic_ga import BasicGA
from .nsga3 import NSGA3

# Deep learning models
from .conv6 import Conv6
from .hwy_blstm import HwyBLSTM
from .lstm import LSTM
from .mlp import MLP

# Extra layers and utils for TF
from .highway_layer import HighwayLayer
from .tf_epoch_logger import TFEpochLogger



#!/usr/bin/python
"""
Non-dominated Sorting Genetic Algorithm with multiple objectives. Based on the
implementation by Tatar et al. [1]_
"""

import random

import numpy as np
from tqdm import tqdm
from deap import algorithms
from deap import base
from deap import creator
from deap import tools

from spiegelib.evaluation.evaluation_base import EvaluationBase
from spiegelib.estimator.estimator_base import EstimatorBase
from spiegelib.synth.synth_base import SynthBase
from spiegelib.features.features_base import FeaturesBase


class NSGA3(EstimatorBase):
    """
        Args:
            synth (Object): Instance of :class:`spiegelib.synth.SynthBase`
            features (list): A list of :class:`spiegelib.features.FeaturesBase` objects.
                Each feature extraction object defines an objective and is used
                in the evaluation function to determine *fitness* of an individual.
            seed (int, optional): Seed for random. Defaults to current system time.
    """

    def __init__(self, synth, features_list, seed=None):
        """
        Constructor
        """
        super().__init__()

        if not isinstance(synth, SynthBase):
            raise TypeError("synth must be of type SynthBase")

        self.synth = synth
        self.num_params = len(synth.get_patch())

        if not isinstance(features_list, list):
            raise TypeError("features_list must be a list")

        self.features_list = features_list
        self.target = None

        self.logbook = tools.Logbook()

        random.seed(seed)
        self._setup()


    def _setup(self):
        """
        Setup genetic algorithm
        """

        self.num_objectives = len(self.features_list)

        # Create individual types
        creator.create("FitnessMin", base.Fitness,
                       weights=(-1.0,) * self.num_objectives)
        creator.create("Individual", list, fitness=creator.FitnessMin)

        # Setup toolbox
        self.toolbox = base.Toolbox()

        # Attribute generator
        self.toolbox.register("attr_float", random.random)

        # Structure initializers
        self.toolbox.register(
            "individual",
            tools.initRepeat,
            creator.Individual,
            self.toolbox.attr_float,
            self.num_params
        )
        self.toolbox.register("population", tools.initRepeat,
                              list, self.toolbox.individual)

        ref_points = tools.uniform_reference_points(self.num_objectives, 12)

        self.toolbox.register("evaluate", self.fitness)
        self.toolbox.register("mate", tools.cxSimulatedBinaryBounded, low=0.0,
                              up=1.0, eta=30.0)
        self.toolbox.register("mutate", tools.mutPolynomialBounded, low=0.0,
                              up=1.0, eta=20.0, indpb=1.0/self.num_params)
        self.toolbox.register("select", tools.selNSGA3, ref_points=ref_points)


    def fitness(self, individual):
        """
        This is automatically called during prediction. Evaluation that calculates
        the fitness of an individual. The individual is a new estimated synthesizer
        parameter setting, and fitness is calculated by rendering an audio sample using
        those parameter settings and then measuring the error between that sample
        and the target sound using a set of audio feature extractors set during
        construction.

        Args:
            individual (list): List of float values representing a synthesizer patch

        Returns:
            list: A list of the error values, one for each feature extractor
        """

        self.synth.set_patch(individual)
        self.synth.render_patch()
        out = self.synth.get_audio()

        errors = []
        index = 0
        for extractor in self.features_list:
            out_features = extractor(out)
            errors.append(EvaluationBase.mean_abs_error(self.target[index],
                                                        out_features))
            index += 1

        return errors


    def predict(self, input):
        """
        Run prection on input audio target

        Args:
            input (:ref:`AudioBuffer <audio_buffer>`): AudioBuffer to use as target
        """

        self.target = []
        for extractor in self.features_list:
            self.target.append(extractor(input))

        pop = self.toolbox.population(n=300)
        invalid_ind = [ind for ind in pop if not ind.fitness.valid]
        fitnesses = self.toolbox.map(self.toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean, axis=0)
        stats.register("std", np.std, axis=0)
        stats.register("min", np.min, axis=0)
        stats.register("max", np.max, axis=0)

        self.logbook.header = "gen", "evals", "std", "min", "avg", "max"

        record = stats.compile(pop)
        self.logbook.record(gen=0, evals=len(invalid_ind), **record)

        # Begin the generational process
        pbar = tqdm(range(1, 100), desc="Generation 1")
        for gen in pbar:
            offspring = algorithms.varAnd(pop, self.toolbox, 0.5, 0.5)

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            fitnesses = self.toolbox.map(self.toolbox.evaluate, invalid_ind)
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit

            # Select the next generation population from parents and offspring
            pop = self.toolbox.select(pop + offspring, 300)

            # Compile statistics about the new population
            record = stats.compile(pop)
            self.logbook.record(gen=gen, evals=len(invalid_ind), **record)
            pbar.set_description("Generation %s" % gen)

        return tools.selBest(pop, 1)[0]



#!/usr/bin/env python
"""
Abstract Base Class for Synthesizer Parameter Estimators
"""

from abc import ABC, abstractmethod

class EstimatorBase(ABC):

    def __init__(self):
        """
        Constructor
        """
        pass


    @abstractmethod
    def predict(self):
        """
        Predict method must be implemented and should estimate parameters
        given some input
        """
        pass
